
// All programs in this collection are free software: 
// you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// Copyright 2007-2008 Alexander Zien.
// Max Planck Intitute for Biological Cybernetics, Tbingen, Germany.
// Friedrich Miescher Lab of the Max Planck Society, Tbingen, Germany.


#include "mex.h"
#include <assert.h>
#include <math.h>



static const char* const copyright =
  "Compute Hessian for Multi-Class Multiple Kernel Learning.";

static const char* const syntax =
  "USAGE: [ MK ] = computeHessian( Kxs, Ky, y )\n";



//=============================================================================//
//=== Auxilliary
//=============================================================================//

//-----------------------------------------------------------------------------//
//--- Matrix Access
//-----------------------------------------------------------------------------//

inline double getKy( const double* const Ky, const int nofClasses, const int i, const int j )
{
  assert( 0 <= i && i < nofClasses );
  assert( 0 <= j && j < nofClasses );
  return( Ky[ i + nofClasses*j ] );
}

inline double getKxs( const double* const Kxs, const int nofKernels, const int m,
		      const int i, const int j, const int p )
{
  assert( 0 <= i && i < m );
  assert( 0 <= j && j < m );
  assert( 0 <= p && p < nofKernels );
  return( Kxs[ i + m*j + m*m*p ] );
}

inline double getKyp( const double* const Kyp, const int nofClasses,
		      const int yip, const int yjp, const int yi, const int yj )
{
  const int nofClasses1 = nofClasses - 1;
  assert( 0 <= yip && yip < nofClasses1 );
  assert( 0 <= yjp && yjp < nofClasses1 );
  assert( 0 <= yi && yi < nofClasses );
  assert( 0 <= yj && yj < nofClasses );
  return( Kyp[ yip + nofClasses1*( yjp + nofClasses1 * ( yi + nofClasses * yj ) ) ] );
}



//=============================================================================//
//=== Main
//=============================================================================//

void mexFunction( int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[] )
{
  // === check number of params
  // - output params
  if( !( 1 <= nlhs && nlhs <= 1 ) ) {
    mexErrMsgTxt( syntax );
  }
  // - input params
  if( !( 3 <= nrhs && nrhs <= 3 ) ) {
    if( nrhs == 0 ) {
      printf( "%s\n", copyright );
      plhs[0] = mxCreateDoubleMatrix( 1, 1, mxREAL );
      double* const results = mxGetPr( plhs[0] );
      results[ 0 ] = mxGetNaN();
      return;
    }
    mexErrMsgTxt( syntax );
  }

  // === process input
  #define Kxs_ ( prhs[ 0 ] )
  #define Ky_ ( prhs[ 1 ] )
  #define y_ ( prhs[ 2 ] )
  // --- initialize variables
  const double* const Kxs = mxGetPr( Kxs_ );
  const double* const Ky = mxGetPr( Ky_ );
  const double* const y = mxGetPr( y_ );
  const int m = mxGetM( y_ );
  const int nofClasses = mxGetM( Ky_ );
  const int dimKxs = mxGetNumberOfDimensions( Kxs_ );
  const int nofKernels = (dimKxs==3) ? mxGetDimensions(Kxs_)[2] : 1;
  // --- check input
  if( !( dimKxs == 3 || dimKxs == 2 ) ) {
    printf( "'Kxs' must be 3-dimensional (or 2-dimensional).\n" );
    mexErrMsgTxt( syntax );
  }
  if( !( mxGetDimensions(Kxs_)[0] == m && mxGetDimensions(Kxs_)[1] == m ) ) {
    printf( "'Kxs' must be of size m*m*nofKernels.\n" );
    mexErrMsgTxt( syntax );
  }
  if( !( mxGetN(Ky_) == nofClasses ) ) {
    printf( "'Ky' must be of size nofClasses square.\n" );
    mexErrMsgTxt( syntax );
  }
  if( !( mxGetN( y_ ) == 1 ) ) {
    printf( "'y' must be a column vector.\n" );
    mexErrMsgTxt( syntax );
  }

  // === compute and return result
  const int nofClasses1 = nofClasses - 1;
  // --- Ky
  // Kyp = zeros(k1,k1,k,k);
  // for( yi = 1:k )
  //   zi = 0;
  //   for( yip = 1:k )
  //     if( yip == yi )
  //       continue;
  //     end;
  //     zi = zi + 1;
  //     for( yj = 1:k )
  //       zj = 0;
  //       for( yjp = 1:k )
  //         if( yjp == yj )
  //           continue;
  //         end;
  //         zj = zj + 1;
  //         Kyp(zi,zj,yi,yj) = Ky(yi,yj)+ Ky(yip,yjp)- Ky(yi,yjp)- Ky(yip,yj);
  //       end;
  //     end;
  //   end;
  // end;
  // assert( size(Kyp) == [k1,k1,k,k] );
  double* Kyp = new double[ nofClasses1*nofClasses1*nofClasses*nofClasses ];
  int yi;
  int yj;
  int yip;
  int yjp;
  register int l;
  l = 0;
  for( yj = 0; yj < nofClasses; ++yj ) {
    for( yi = 0; yi < nofClasses; ++yi ) {
      for( yjp = 0; yjp < nofClasses; ++yjp ) {
	if( yjp == yj ) {
	  continue;
	}
	for( yip = 0; yip < nofClasses; ++yip ) {
	  if( yip == yi ) {
	    continue;
	  }
	  // Kyp[l] == Kyp[yip,yjp,yi,yj]
	  Kyp[l] = getKy(Ky,nofClasses,yi,yj) + getKy(Ky,nofClasses,yip,yjp) - getKy(Ky,nofClasses,yi,yjp) - getKy(Ky,nofClasses,yip,yj);
	  ++l;
	}
      }
    }
  }
  assert( l == nofClasses1*nofClasses1*nofClasses*nofClasses );
  // --- MK
  // MK = zeros(m*k1,m*k1,nofKernels);
  // for( p = 1:nofKernels )
  //   for( i = 1:m )
  //     idx_i = (k1*(i-1)+1) : (k1*i);
  //     for( j = 1:m )
  //       idx_j = (k1*(j-1)+1) : (k1*j);
  //       MK(idx_i,idx_j,p) = Kyp(:,:,y(i),y(j)) * Kx(i,j,p);
  //     end;
  //   end;
  // end;
  const int dims[] = { m*nofClasses1, m*nofClasses1, nofKernels };
  plhs[0] = mxCreateNumericArray( 3, dims, mxDOUBLE_CLASS, mxREAL );
  double* const MK = mxGetPr( plhs[0] );
  int p;
  int i;
  int j;
  int k;
  l = 0;
  k = 0;
  for( p = 0; p < nofKernels; ++p ) {
    for( j = 0; j < m; ++j ) {
      const int yj = (int) y[ j ];
      assert( 0 <= yj && yj < nofClasses );
      for( yjp = 0; yjp < nofClasses1; ++yjp ) {
	for( i = 0; i < m; ++i ) {
	  const int yi = (int) y[ i ];
	  const double Kxs_ijp = getKxs( Kxs, nofKernels, m, i, j, p );
	  for( yip = 0; yip < nofClasses1; ++yip ) {
	    // MK[l] == MK[yip,i,yjp,j,p]
	    MK[l] = getKyp(Kyp,nofClasses,yip,yjp,yi,yj) * Kxs_ijp;
	    ++l;
	  }
	}
      }
    }
  }
  assert( l == m*nofClasses1 * m*nofClasses1 * nofKernels );
  delete[] Kyp;
}




