
% All programs in this collection are free software: 
% you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright 2007-2008 Alexander Zien.
% Max Planck Intitute for Biological Cybernetics, Tbingen, Germany.
% Friedrich Miescher Lab of the Max Planck Society, Tbingen, Germany.


%==============================================================================
function [ svm, param, resQP ] = mcSvmTrainCompr( param, lastSvm, Kx, y, idxTrain );
%==============================================================================


%==============================================================================
% Init
%==============================================================================

% === init
mAll = size( y, 1 );
classes = unique( y )';
nofClasses = length( classes );
nofClasses1 = nofClasses - 1;
if( ~exist('idxTrain','var') )
  idxTrain = 1:mAll;
end;
m = length( idxTrain );

% === check
% --- dims
ASSERT( size(y) == [mAll,1] );
ASSERT( size(Kx) == [mAll,mAll] );
ASSERT( isscalar(param.C) );
% --- content
ASSERT( isfinite(Kx) );
ASSERT( classes == (1:nofClasses)-1 );
ASSERT( 0 < param.C );

% === params
param = setDefault( param, 'verbosity', 0 );
% --- model
param = setDefault( param, 'Ky', eye(nofClasses) );
param = setDefault( param, 'rescaling', 'none' );  % 'none', 'margin', 'slope'
param = setDefault( param, 'classWeights', [] );   % only required for rescaling
% --- optimization
param = setDefault( param, 'epsAlpha', 1e-12 );
param = setDefault( param, 'epsC', 1e-8 );
% --- optimization
param = setDefault( param, 'optParam', [] );
optParam = param.optParam;
optParam = setDefault( optParam, 'optimizer', 'cplexgun' );  % 'toolbox', 'cplexint'
optParam = setDefault( optParam, 'cplexQpSolver', 'net' );

% === check params
ASSERT( size(param.Ky) == [nofClasses,nofClasses] );
ASSERT( isfinite(param.Ky) );


%==============================================================================
% Prepare QP
%==============================================================================

% === objective
fqp = -ones( m*nofClasses1, 1 );
H = computeHessian( Kx(idxTrain,idxTrain), param.Ky, y(idxTrain) );
%H = H + 1e-10*speye(size(H));  % hack to ensure positive semidefiniteness
ASSERT( size(H) == [m*nofClasses1,m*nofClasses1] );

% === constraints
Aqp = zeros( nofClasses+m, m*nofClasses1 );
for( k = 1:nofClasses )
  for( i = 1:m )
    y_i = y(idxTrain(i)) + 1;
    if( y_i == k )
      Aqp( k, (i-1)*nofClasses1 + (1:nofClasses1) ) = -1;
    else
      k1 = k - ( y_i < k );
      Aqp( k, (i-1)*nofClasses1 + k1 ) = 1;
    end;
  end;
end;
for( i = 1:m )
  Aqp( nofClasses+i, (i-1)*nofClasses1 + (1:nofClasses1) ) = 1;
end;
if( ~strcmp(optParam.optimizer,'toolbox') )
  Aqp = sparse( Aqp );
end;
bqp = [ zeros(nofClasses,1); repmat( param.C, m, 1 ) ];
neqqp = nofClasses;
LBqp = zeros( m*nofClasses1, 1 );
UBqp = repmat( inf, m*nofClasses1, 1 );

% === rescaling
if( param.verbosity >= 1 )
  fprintf( 'rescaling: %s\n', param.rescaling );
end;
switch( param.rescaling )
 case 'none'
  ASSERT( isempty(param.classWeights) );
 case 'margin'
  ASSERT( size(param.classWeights) == [1,nofClasses] );
  for( i = 1:m )
    fqp( (i-1)*nofClasses1 + (1:nofClasses1) ) = -param.classWeights( y(idxTrain(i))+1 );
  end;
 case 'slope'
  ASSERT( size(param.classWeights) == [1,nofClasses] );
  for( i = 1:m )
    bqp( i ) = C * param.classWeights( y(idxTrain(i))+1 );
  end;
 otherwise
  error( 'unknown rescaling' );
end;

% === MP object
QP = [];
QP.type = 'QP';
QP.verbosity = 1;
QP.c = fqp;
QP.Q = sparse( H );
QP.A = Aqp;
QP.b = bqp;
QP.neq = neqqp;
QP.LB = LBqp;
QP.UB = UBqp;


%==============================================================================
% Solve QP
%==============================================================================

% === start value
if( isempty(lastSvm) )
  alphasCompressed = zeros( m*nofClasses1, 1 );
  alphasCompressed(1) = param.C;
  alphasCompressed = reshape( alphasCompressed, nofClasses1, m );
else
  alphasCompressed = lastSvm.alphasCompressed;
end;

% === solve QP
[ optParam ] = MP_getLicence( optParam );
[ alphasCompressed, resQP ] = MP_solve( optParam, QP, alphasCompressed(:) );
[ optParam ] = MP_releaseLicence( optParam );
alphasCompressed( alphasCompressed < param.epsAlpha ) = 0;
regul = alphasCompressed' * H * alphasCompressed;
alphasCompressed = reshape( alphasCompressed, nofClasses1, m );

% === compute output differences on free SVs
sa = sum( alphasCompressed', 2 );
idxSvs = find( param.epsC*param.C<sa & sa<param.C*(1-param.epsC) );
nofSvs = length( idxSvs );
idxSvAlphas = repmat((idxSvs-1)*nofClasses1,1,nofClasses1) + repmat(1:nofClasses1,nofSvs,1);
idxSvAlphas = idxSvAlphas';
idxSvAlphas = idxSvAlphas(:);
outDiffsSvs = H(idxSvAlphas,:) * alphasCompressed(:);
outDiffsSvs = [ reshape(outDiffsSvs,nofClasses1,nofSvs)' , zeros(nofSvs,1) ];
for( i = 1:nofSvs )
  yi = y(idxTrain(idxSvs(i))) + 1;
  outDiffsSvs(i,(yi+1):nofClasses) = outDiffsSvs(i,yi:nofClasses1);
  outDiffsSvs(i,yi) = 0;
end;
clear H;


%==============================================================================
% Post-Processing
%==============================================================================

% === unfold alphas
alphasUnfolded = [ alphasCompressed' , zeros(m,1) ];
for( i = 1:m )
  yi = y(idxTrain(i)) + 1;
  alphasUnfolded(i,(yi+1):nofClasses) = alphasUnfolded(i,yi:nofClasses1);
  alphasUnfolded(i,yi) = 0;
end;

% === compute biases
A = [];
b = [];
for( k = 1:length(idxSvs) )
  i = idxSvs(k);
  for( u = find( alphasUnfolded(i,:) ) )
    a = zeros( 1, nofClasses );
    a( y(idxTrain(i))+1 ) = 1;
    a( u ) = -1;
    A = [ A ; a ];
    b = [ b ; 1-outDiffsSvs(k,u) ];
  end;
end;
if( nofSvs < nofClasses1 )
  biases = zeros( 1, nofClasses );
else
  biases = ( pinv(A) * b )';
  biases = biases - mean(biases);
end;

% === convert alphas
alphasUnfolded = -alphasUnfolded;
for( i = 1:m )
  yi = y(idxTrain(i)) + 1;
  ASSERT( alphasUnfolded(i,yi) == 0 );
  alphasUnfolded(i,yi) = - sum( alphasUnfolded(i,:) );
end;

% === return results
svm = [];
svm.y = y(idxTrain);
svm.alphasCompressed = alphasCompressed;
svm.alphas = alphasUnfolded;
svm.biases = biases;
svm.sumOfAlphas = -fqp' * alphasCompressed(:);
svm.regul = regul;
svm.objVal = resQP.objVal;
ASSERT( abs( svm.objVal - (svm.regul-svm.sumOfAlphas) ) < 1e8 );

